/*
 * Copyright (c) 2017 SpaceToad and the BuildCraft team
 * This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0. If a copy of the MPL was not
 * distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/
 */

package buildcraft.transport.pipe;

import java.io.IOException;

import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.network.PacketBuffer;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.ResourceLocation;

import net.minecraftforge.fml.common.network.simpleimpl.MessageContext;
import net.minecraftforge.fml.relauncher.Side;

import buildcraft.api.core.BCLog;
import buildcraft.api.transport.pipe.PipeApi;
import buildcraft.api.transport.pluggable.PipePluggable;
import buildcraft.api.transport.pluggable.PluggableDefinition;

import buildcraft.lib.net.PacketBufferBC;

import buildcraft.transport.tile.TilePipeHolder;

public final class PluggableHolder {
    private static final int ID_REMOVE_PLUG = 0;
    private static final int ID_UPDATE_PLUG = 1;
    private static final int ID_CREATE_PLUG = 2;

    public final TilePipeHolder holder;
    public final EnumFacing side;
    public PipePluggable pluggable;
    /** Used to determine if a full "create" message should be sent when {@link #writePayload(PacketBufferBC, Side)} is
     * called. If this is false it means that last time it was null, and a create message should. */
    private boolean lastGeneralExisted = false;

    public PluggableHolder(TilePipeHolder holder, EnumFacing side) {
        this.holder = holder;
        this.side = side;
    }

    // Saving + Loading

    public NBTTagCompound writeToNbt() {
        NBTTagCompound nbt = new NBTTagCompound();
        if (pluggable != null) {
            nbt.func_74778_a("id", pluggable.definition.identifier.toString());
            nbt.func_74782_a("data", pluggable.writeToNbt());
        }
        return nbt;
    }

    public void readFromNbt(NBTTagCompound nbt) {
        if (nbt.func_82582_d()) {
            pluggable = null;
            return;
        }
        String id = nbt.func_74779_i("id");
        NBTTagCompound data = nbt.func_74775_l("data");
        ResourceLocation identifier = new ResourceLocation(id);
        PluggableDefinition def = PipeApi.pluggableRegistry.getDefinition(identifier);
        if (def == null) {
            BCLog.logger.warn("Unknown pluggable id '" + id + "'");
            throw new Error("Def was null!");
        } else {
            pluggable = def.readFromNbt(holder, side, data);
            holder.eventBus.registerHandler(pluggable);
            lastGeneralExisted = true;
        }
    }

    // Network

    public void writeCreationPayload(PacketBuffer buffer) {
        if (pluggable == null) {
            buffer.writeByte(ID_REMOVE_PLUG);
        } else {
            buffer.writeByte(ID_CREATE_PLUG);
            buffer.func_180714_a(pluggable.definition.identifier.toString());
            pluggable.writeCreationPayload(buffer);
        }
    }

    public void readCreationPayload(PacketBuffer buffer) {
        int id = buffer.readUnsignedByte();
        if (id == ID_CREATE_PLUG) {
            readCreateInternal(buffer);
        } else {
            holder.eventBus.unregisterHandler(pluggable);
            pluggable = null;
        }
    }

    private void readCreateInternal(PacketBuffer buffer) {
        ResourceLocation identifier = new ResourceLocation(buffer.func_150789_c(256));
        PluggableDefinition def = PipeApi.pluggableRegistry.getDefinition(identifier);
        if (def == null) {
            throw new IllegalStateException("Unknown remote pluggable \"" + identifier + "\"");
        }
        pluggable = def.loadFromBuffer(holder, side, buffer);
        holder.eventBus.registerHandler(pluggable);
    }

    public void writePayload(PacketBufferBC buffer, Side netSide) {
        if (netSide == Side.CLIENT) {
            buffer.writeByte(ID_UPDATE_PLUG);
            if (pluggable != null) {
                pluggable.writePayload(buffer, netSide);
            }
        } else {
            if (pluggable == null) {
                lastGeneralExisted = false;
                buffer.writeByte(ID_REMOVE_PLUG);
            } else if (lastGeneralExisted) {
                buffer.writeByte(ID_UPDATE_PLUG);
                pluggable.writePayload(buffer, netSide);
            } else {
                // The last general one did NOT exist, and so we need to create it
                lastGeneralExisted = true;
                writeCreationPayload(buffer);
            }
        }
    }

    public void readPayload(PacketBufferBC buffer, Side netSide, MessageContext ctx) throws IOException {
        int id = buffer.readUnsignedByte();
        if (netSide == Side.SERVER) {
            if (id == ID_UPDATE_PLUG) {
                if (pluggable != null) {
                    pluggable.readPayload(buffer, netSide, ctx);
                }
            } else {
                BCLog.logger.warn("[PluggableHolder] Unknown ID " + id);
            }
        } else {
            if (id == ID_REMOVE_PLUG) {
                holder.eventBus.unregisterHandler(pluggable);
                pluggable = null;
            } else if (id == ID_UPDATE_PLUG) {
                pluggable.readPayload(buffer, netSide, ctx);
            } else if (id == ID_CREATE_PLUG) {
                readCreateInternal(buffer);
            } else {
                BCLog.logger.warn("[PluggableHolder] Unknown ID " + id);
            }
        }
    }

    // Pluggable overrides

    public void onTick() {
        if (pluggable != null) {
            pluggable.onTick();
        }
    }
}
